/*
 * Copyright (C) 2002 2003 Sven Schaepe <schaepe@rz.tu-ilmenau.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include <string.h>

#include "utils.h"
#include "desc.h"

FILE *OUTPUT;
int DEBUG=0;
char LOOP;

typedef struct {
  int stable;     //standart tabelle
  int ctable;     //compare tabelle
} tINDEX;

typedef struct {
  int descriptor_tag;
  const char *name;
  sTABLE *table;
  unsigned int length;
} sDESCRIPTOR_ID;

sTABLE v_s_Desc[] = {
  { "multiple_frame_rate_flag",       1, 0 , 0 },
  { "frame_rate_code",                4, 0 , 0 },
  { "MPEG_1_only_flag",               1, 0 , 0 },
  { "constrained_parameter_flag",     1, 0 , 0 },
  { "still_picture_flag",             1, 0 , 0 },
  { "d_loop_1",                       0,-1 , VS_loop }
};

sTABLE v_s_Descl[] = {
  { "profile_and_level_indication",   8, 0 , 0 },
  { "chroma_format",                  2, 0 , 0 },
  { "frame_rate_extension_flag",      1, 0 , 0 },
  { "reserved",                       5, 0 , 0 }
};

sTABLE a_s_Desc[] = {
  { "free_format_flag",               1, 0 , 0 },
  { "ID",                             1, 0 , 0 },
  { "layer",                          2, 0 , 0 },
  { "variable_rate_audio_indicator",  1, 0 , 0 },
  { "reserved",                       3, 0 , 0 }
};

sTABLE h_Desc[] = {
  { "reserved",                       4, 0 , 0 },
  { "hierarchy_type",                 4, 0 , 0 },
  { "reserved",                       2, 0 , 0 },
  { "hierarchy_layer_index",          6, 0 , 0 },
  { "reserved",                       2, 0 , 0 },
  { "hierarchy_embedded_layer_index", 6, 0 , 0 },
  { "reserved",                       2, 0 , 0 },
  { "hierarchy_channel",              6, 0 , 0 }
};  

sTABLE d_s_a_Desc[] = {
  { "alignment_type",                 8, 0 , 0 }
};  

sTABLE t_b_g_Desc[] = {
  { "horizontal_size",               14, 0 , 0 },
  { "vertical_size",                 14, 0 , 0 },
  { "aspect_ratio_information",       4, 0 , 0 }
};  

sTABLE v_w_Desc[] = {
  { "horizontal_offset",             14, 0 , 0 },
  { "vertical_offset",               14, 0 , 0 },
  { "window_priority",                4, 0 , 0 }
};  

sTABLE ca_Desc[] = {
  { "CA_system_ID",                  16, 0 , 0 },
  { "reserved",                       3, 0 , 0 },
  { "CA_PID",                        13, 0 , 0 },
  { "d_loop",                         0,-2 , DESC_loop_end_memcpy }
};  

sTABLE ca_Descl[] = {
  { "private_data_byte",              8, 0 , 0 }
};

sTABLE i_l_Desc[] = {
  { "d_loop",                         0,-3 , DESC_loop_end }
};

sTABLE i_l_Descl[] = {
  { "ISO_639_language_code",         24, 0 , STRING2 },
  { "audio_type",                     8, 0 , 0 }
};

sTABLE s_c_Desc[] = {
  { "external_clock_reference_indicator",  1, 0 , 0 },
  { "reserved",                            1, 0 , 0 },
  { "clock_accuracy_integer",              6, 0 , 0 },
  { "clock_accuracy_exponent",             3, 0 , 0 },
  { "reserved",                            5, 0 , 0 }
};  

sTABLE m_b_u_Desc[] = {
  { "bound_valid_flag",               1, 0 , 0 },
  { "LTW_offset_lower_bound",        15, 0 , 0 },
  { "reserved",                       1, 0 , 0 },
  { "LTW_offset_upper_bound",        15, 0 , 0 }
};  

sTABLE cr_Desc[] = {
  { "copyright_identifier",          32, 0 , 0 },
  { "d_loop",                         0,-4 , DESC_loop_end }
};

sTABLE cr_Descl[] = {
  { "additional_copyright_info",      8, 0 , 0 }
};
 
sTABLE m_b_Desc[] = {
  { "reserved",                       2, 0 , 0},
  { "maximum_bitrate",               22, 0 , 0}
};  

sTABLE p_d_i_Desc[] = {
  { "private_data_indicator",        32, 0 , 0 }
};  

sTABLE s_b_Desc[] = {
  { "reserved",                       2, 0 , 0 },
  { "sb_leak_rate",                  22, 0 , 0 },
  { "reserved",                       2, 0 , 0 },
  { "sb_size",                       22, 0 , 0 }
};  

sTABLE sdt_Desc[] = {
  { "reserved",                       7, 0 , 0 },
  { "leak_valid_flag",                1, 0 , 0 }
};  

sTABLE ibp_Desc[] = {
  { "closed_gop_flag",                1, 0 , 0 },
  { "identical_gop_flag",             1, 0 , 0 },
  { "max_gop-length",                14, 0 , 0 }
};  

sTABLE n_n_Desc[] = {
  { "string",                         0, 0, STRING1 }
};  

sTABLE s_l_Desc[] = {
  { "d_loop",                         0,-6 , DESC_loop_end }
};  

sTABLE s_l_Descl[] = {
  { "service_id",                    16, 0 , 0 },
  { "service_type",                   8, 0 , 0 }
};  

sTABLE s_d_s_Desc[] = {
  { "frequency",                     32, 0 , 0 },
  { "orbital_position",              16, 0 , 0 },
  { "west_east_flag",                 1, 0 , 0 },
  { "polarisation",                   2, 0 , 0 },
  { "modulation",                     5, 0 , 0 },
  { "symbolrate",                    28, 0 , 0 },
  { "FEC_inner",                      4, 0 , 0 }
};  

sTABLE b_n_Desc[] = {
  { "string",                         0, 0, STRING1 }
};  

sTABLE s_Desc[] = {
  { "service_type",                   8, 0 , 0 },
  { "service_provider_name_length",   8, 0 , 0 },
  { "string",                         0, 0, STRING },
  { "service_name_length",            8, 0 , 0 },
  { "string",                         0, 0, STRING }
};  

sTABLE c_a_Desc[] = {
  { "country_availability_flag",      1, 0 , 0 },
  { "reserved_future_use",            7, 0 , 0 },
  { "d_loop",              	      0,-14 , DESC_loop_end }
};  

sTABLE c_a_Descl[] = {
  { "country_code",   		     24, 0 , STRING2 },
};

sTABLE l_Desc[] = {
  { "transport_stream_id",            16, 0 , 0 },
  { "orginal_network_id",             16, 0 , 0 },
  { "service_id",                     16, 0 , 0 },
  { "linkage_type",                    8, 0 , 0 },
  { "linkage_type",                    0,-8 , LINKAGE }
};  

sTABLE l_Descl[] = {   //linkage_type!=0x8
  { "string",                          0, 0, STRING1 }
};  

sTABLE l_Descl1[] = {   //linkage_type==0x8
  { "hand_over_type",                  4, 0 , 0 },
  { "reserved_future_use",             3, 0 , 0 },
  { "origin_type",                     1, 0 , 0 },
  { "network/initial_service_id",     16, 0 , 0 },
  { "string",                          0, 0, STRING }
};  

sTABLE t_s_s_Desc[] = {
  { "referenc_service_id",            16, 0 , 0 }
};

sTABLE NVOD_Descl[] = {
  { "transport_stream_id",            16, 0 , 0 },
  { "orginal_network_id",             16, 0 , 0 },
  { "service_id",                     16, 0 , 0 },
};

sTABLE NVOD_Desc[] = {
  { "d_loop",              	       0,-12 , DESC_loop_end }
};

sTABLE s_e_Desc[] = {
  { "ISO 639-2[3]_language_code",    24, 0 , STRING2 },
  { "event_name_length",              8, 0 , 0 },
  { "string",                         0, 0, STRING },
  { "text_length",                    8, 0 , 0 },
  { "string",                         0, 0, STRING }
};  

sTABLE e_e_Desc[] = {
  { "descriptor_number",              4, 0 , 0 },
  { "last_descriptor_number",         4, 0 , 0 },
  { "ISO 639-2[3]_language_code",    24, 0 , STRING2 },
  { "length_of_item",                 8, 0 , 0 },
  { "d_loop1",                        0,-7 , DESC_loop_kl },
  { "text_length",                    8, 0 , 0 },
  { "string",                         0, 0, STRING }
};  

sTABLE e_e_Descl[] = {
  { "item_descriptor_length",         8, 0 , 0 },
  { "string",                         0, 0, STRING },
  { "item_length",                    8, 0 , 0 },
  { "string",                         0, 0, STRING }
};  

sTABLE t_s_e_Desc[] = {
  { "reference_service_id",          16, 0 , 0 },
  { "reference_event_id",            16, 0 , 0 }
};

sTABLE com_Desc[] = {
  { "reserved_future_use",            4, 0 , 0 },
  { "stream_content",                 4, 0 , 0 },
  { "component_type",                 8, 0 , 0 },
  { "component_tag",                  8, 0 , 0 },
  { "ISO 639-2 [3]_language_code",   24, 0 , STRING2 },
  { "string",                         0, 0, STRING1 }
};  

sTABLE s_i_Desc[] = {
  { "component_tag",                  8, 0 , 0 }
};  

sTABLE ca_i_Desc[] = {
  { "d_loop",                        0,-10 , DESC_loop_end }
};  

sTABLE ca_i_Descl[] = {
  { "CA_system_id",                   16, 0 , 0 }
};  

sTABLE c_Desc[] = {
  { "d_loop",                         0,-5 , DESC_loop_end }
};  

sTABLE c_Descl[] = {
  { "c_n_l_1",                        4, 0 , 0 },
  { "c_n_l_2",                        4, 0 , 0 },
  { "u_n_1",                          4, 0 , 0 },
  { "u_n_2",                          4, 0 , 0 }
};  

sTABLE p_r_Desc[] = {
  { "d_loop",                         0,-13 , DESC_loop_end }
};  

sTABLE p_r_Descl[] = {
  { "ISO 639-2 [3]_language_code",   24, 0 , STRING2 },
  { "rating",                         8, 0 , 0 },
};

sTABLE st_Desc[] = {
  { "d_loop",                         0,-15 , DESC_loop_end }
};

sTABLE st_Descl[] = {
  { "ISO 639-2[3]_language_code",     24, 0, STRING2 },
  { "subtitle_type",                   8, 0 , 0 },
  { "composition_page_id",            16, 0 , 0 },
  { "ancilary_page_id",               16, 0 , 0 },
};

sTABLE tt_Desc[] = {
};  

sTABLE tt_Descl[] = {
  { "ISO 639-2[3]_language_code",     24, 0, STRING2 },
  { "teletext_type",                   5, 0, 0 },
  { "teletext_magazin_number",         3, 0, 0 },
  { "teletext_page_number",            8, 0, 0 }
};  

sTABLE p_d_s_Desc[] = {
  { "private_data_specifier",        32, 0 , 0 }
};  

sTABLE PDC_Desc[] = {
  { "reserved_future_use",            4, 0 , 0 },
  { "date,time",                     20, 0 , 0 }
//  { "day",                            5, 0 , 0 },
//  { "month",                          4, 0 , 0 },
//  { "hour",                           5, 0 , 0 },
//  { "minute",                         6, 0 , 0 }
};  

sDESCRIPTOR_ID descriptorId[] = {
  {-15   , "subtitling_desc_loop",           st_Descl   , TABLESIZE(st_Descl)   },
  {-14   , "country_availability_desc_loop", c_a_Descl  , TABLESIZE(c_a_Descl)  },
  {-13   , "parental_rating_desc_loop",      p_r_Descl  , TABLESIZE(p_r_Descl)  },
  {-12   , "NVOD_ref_desc_loop",             NVOD_Descl , TABLESIZE(NVOD_Descl) },
  {-11   , "teletext_desc_loop",             tt_Descl   , TABLESIZE(tt_Descl)   },
  {-10   , "CA_identifier_desc_loop",        ca_i_Descl , TABLESIZE(ca_i_Descl) },
  {-9    , "linkage_desc==8",                l_Descl1   , TABLESIZE(l_Descl1)   },
  {-8    , "linkage_desc!=8",                l_Descl    , TABLESIZE(l_Descl)    },
  {-7    , "extended_event_desc_loop",       e_e_Descl  , TABLESIZE(e_e_Descl)  },
  {-6    , "service_list_desc_loop",         s_l_Descl  , TABLESIZE(s_l_Descl)  },
  {-5    , "content_desc_loop",              c_Descl    , TABLESIZE(c_Descl)    },
  {-4    , "copyright_desc_loop",            cr_Descl   , TABLESIZE(cr_Descl)   },
  {-3    , "ISO_639_language_desc_loop",     i_l_Descl  , TABLESIZE(i_l_Descl)  },
  {-2    , "CA_desc_loop",                   ca_Descl   , TABLESIZE(ca_Descl)   },
  {-1    , "MPEG_1_only_flag=1",             v_s_Descl  , TABLESIZE(v_s_Descl)  },
  { 2    , "video_stream_desc" ,             v_s_Desc   , TABLESIZE(v_s_Desc)   },
  { 3    , "audio_stream_desc" ,             a_s_Desc   , TABLESIZE(a_s_Desc)   },
  { 4    , "hierarchy_desc" ,                h_Desc     , TABLESIZE(h_Desc)     },
  { 6    , "data_stream_alignment_desc" ,    d_s_a_Desc , TABLESIZE(d_s_a_Desc) },
  { 7    , "target_background_grid_desc" ,   t_b_g_Desc , TABLESIZE(t_b_g_Desc) },
  { 8    , "video_window_desc" ,             v_w_Desc   , TABLESIZE(v_w_Desc)   },
  { 9    , "CA_desc" ,                       ca_Desc    , TABLESIZE(ca_Desc)    },
  { 10   , "ISO_639_language_desc" ,         i_l_Desc   , TABLESIZE(i_l_Desc)   },
  { 11   , "system_clock_desc" ,             s_c_Desc   , TABLESIZE(s_c_Desc)   },
  { 12   , "multi_buffer_utilization_desc" , m_b_u_Desc , TABLESIZE(m_b_u_Desc) },
  { 13   , "copyright_desc" ,                cr_Desc    , TABLESIZE(cr_Desc)    },
  { 14   , "maximum_bitrate_desc" ,          m_b_Desc   , TABLESIZE(m_b_Desc)   },
  { 15   , "privat_data_indicator_desc" ,    p_d_i_Desc , TABLESIZE(p_d_i_Desc) },
  { 16   , "smoothing_buffer_desc" ,         s_b_Desc   , TABLESIZE(s_b_Desc)   },
  { 17   , "SDT_desc" ,                      sdt_Desc   , TABLESIZE(sdt_Desc)   },
  { 18   , "IBP_desc" ,                      ibp_Desc   , TABLESIZE(ibp_Desc)   },
  {0x40  , "network_name_desc",              n_n_Desc   , TABLESIZE(n_n_Desc)   },
  {0x41  , "service_list_desc",              s_l_Desc   , TABLESIZE(s_l_Desc)   },
  {0x43  , "satellite_delivery_system_desc", s_d_s_Desc , TABLESIZE(s_d_s_Desc) },
  {0x47  , "bouquet_name_desc",              b_n_Desc   , TABLESIZE(b_n_Desc)   },
  {0x48  , "service_desc",                   s_Desc     , TABLESIZE(s_Desc)     },
  {0x49  , "country_availability_desc",      c_a_Desc   , TABLESIZE(c_a_Desc)   },
  {0x4a  , "linkage_desc",                   l_Desc     , TABLESIZE(l_Desc)     },
  {0x4b  , "NVOD_ref_desc",                  NVOD_Desc  , TABLESIZE(NVOD_Desc)  },
  {0x4c  , "time_shift_service_desc",        t_s_s_Desc , TABLESIZE(t_s_s_Desc) },
  {0x4d  , "short_event_desc",               s_e_Desc   , TABLESIZE(s_e_Desc)   },
  {0x4e  , "extended_event_desc",            e_e_Desc   , TABLESIZE(e_e_Desc)   },
  {0x4f  , "time_shift_even_desc",           t_s_e_Desc , TABLESIZE(t_s_e_Desc) },
  {0x50  , "component_desc",                 com_Desc   , TABLESIZE(com_Desc)   },
  {0x52  , "stream_identifier_desc",         s_i_Desc   , TABLESIZE(s_i_Desc)   },
  {0x53  , "CA_identifier_desc",             ca_i_Desc  , TABLESIZE(ca_i_Desc)  },
  {0x54  , "content_desc",                   c_Desc     , TABLESIZE(c_Desc)     },
  {0x55  , "parental_rating_desc",           p_r_Desc   , TABLESIZE(p_r_Desc)   },
  {0x56  , "teletext_desc",                  tt_Desc    , TABLESIZE(tt_Desc)    },
  {0x59  , "subtitling_desc",          	     st_Desc    , TABLESIZE(st_Desc)    },
  {0x5f  , "private_data_specifier_desc",    p_d_s_Desc , TABLESIZE(p_d_s_Desc) },
  {0x69  , "PDC_desc",                       PDC_Desc   , TABLESIZE(PDC_Desc)   }
};


void get_string(unsigned char *buffer,unsigned int length,unsigned char *String) 
{
    String[0]='\0';
    memcpy(String,buffer,length);
    String[length]='\0';
}

static tINDEX getdescriptor(int descriptor_tag, cDESC *c_desc, unsigned int c_desc_size) 
{
    tINDEX tables;
    int zeile,k;

    for (zeile=0;zeile<TABLESIZE(descriptorId);zeile++) {
	if(descriptorId[zeile].descriptor_tag==descriptor_tag) { //desc bekannt
	    tables.stable=zeile;
	    if (!ALL) {
		for(k=0;k<c_desc_size;k++) {
		    if(c_desc[k].desc_id==descriptor_tag){
			tables.ctable=k;
			return tables;
		    }
		}
	    }
	    tables.ctable=zeile;//0xffff;
	    return tables;
	}
    }
    tables.stable=0xffff;
    tables.ctable=0xffff;
    return tables;
}
  
static int analyse_descriptor(unsigned char *buffer,unsigned int desc_length
			    ,LIST *todo,unsigned int descriptor_tag
			    ,unsigned int *level,int channelnr,int loop_count) 
{
    int j,subtable;
    tINDEX tables;
    unsigned int pos,len;
    unsigned int d_zeile,stringlength;
    unsigned int last=0;
    unsigned char String[256+1];
    sTABLE *desctable;

    tables=getdescriptor(descriptor_tag,todo->desc_encoding,todo->desc_encoding_size);
    if(tables.stable==0xffff) {
	if(ALL) {
	    /* durch diese 3 zeilen entsteht ein merkwuerdiger bug */
	    /* im fwrite und nur bei der ARD und dem ZDF */
	    /* aufgetreten beim gcc-2.95.3 libc-2.2.5 */
	    tprintf(OUTPUT,*level,"unknown descriptor <%x> hexdump mit %x bytes\n",descriptor_tag,desc_length/8);
	    print_hex_dump(buffer,desc_length/8,level,OUTPUT);
	    tprintf(OUTPUT,*level,"unknown descriptor end <%x> hexdump\n",descriptor_tag);
	}
	return desc_length;
    }
    if(!ALL && tables.ctable==0xffff) {
	return desc_length;
    }
    
    desctable=descriptorId[tables.stable].table;
    if(ALL) {
	tprintf(OUTPUT,*level,"<%s=0x%x> <mit der lnge 0x%x>\n"
		,descriptorId[tables.stable].name,descriptor_tag,desc_length/8);
    }
    (*level)++;
    for(pos=0;pos<desc_length;) {
	for(d_zeile=0;d_zeile<(descriptorId[tables.stable].length);d_zeile++) {    //eintraege pro tabelle
	    switch(desctable[d_zeile].encoding) {
		case STRING2:
		    stringlength=desctable[d_zeile].bit_count;
		    get_string(&buffer[pos/8],stringlength/8,String); 
		    if(ALL) {
			tprintf(OUTPUT,*level,"%s: <\"%s\">\n"
					,desctable[d_zeile].name,String);
		    } else {
			for(j=0;j<todo->desc_encoding_size;j++){
			    if(descriptor_tag==(todo->desc_encoding[j].desc_id)) {
				if((todo->desc_encoding[j].d_zeile)==d_zeile){
				    memcpy(todo->desc_encoding[j].save_in
				    +todo->desc_encoding[j].save_in_struct_size
				    *loop_count
				    /sizeof(todo->desc_encoding[j].save_in)
				    +todo->encoding[0].save_in_struct_size
				    *channelnr
				    /sizeof(todo->encoding[0].save_in) 
				    ,String
				    ,4);
				    LOOP=1;
				    break;
				}
			    }
			}
		    }
		    pos+=stringlength;
		    break;
		case STRING1: 
		    stringlength=desc_length-pos;
		    get_string(&buffer[pos/8],stringlength/8,String); 
		    if(ALL) {
			tprintf(OUTPUT,*level,"%s: <\"%s\">\n"
					,desctable[d_zeile].name,String);
		    } else {
			for(j=0;j<todo->desc_encoding_size;j++){
			    if(descriptor_tag==(todo->desc_encoding[j].desc_id)) {
				if((todo->desc_encoding[j].d_zeile)==d_zeile){
				    memcpy(todo->desc_encoding[j].save_in
				    +todo->desc_encoding[j].save_in_struct_size
				    *loop_count
				    /sizeof(todo->desc_encoding[j].save_in)
				    +todo->encoding[0].save_in_struct_size
				    *channelnr
				    /sizeof(todo->encoding[0].save_in) 
				    ,String
				    ,256);
				    LOOP=1;
				}
			    }
			}
		    }
		    pos+=stringlength;
		    break;
		case STRING: 
		    stringlength=last*8;
		    get_string(&buffer[pos/8],stringlength/8,String); 
		    if(ALL) { 
			printf("\n%s\n\n",String);
		    } else {
			for(j=0;j<todo->desc_encoding_size;j++){
			    if(descriptor_tag==(todo->desc_encoding[j].desc_id)) {
				if((todo->desc_encoding[j].d_zeile)==d_zeile){
				    memcpy(todo->desc_encoding[j].save_in
				    +todo->desc_encoding[j].save_in_struct_size
				    *loop_count
				    /sizeof(todo->desc_encoding[j].save_in)
				    +todo->encoding[0].save_in_struct_size
				    *channelnr
				    /sizeof(todo->encoding[0].save_in) 
				    ,String
				    ,last+1);
				    LOOP=1;
				    break;
				}
			    }
			}
		    }
		    pos+=stringlength;
		    break;
		case DESC_loop_kl:
		    len=last*8;
		    if(len>0){  /* hack, falls ein LOOP 0 byte ist */
		      (*level)--;
		      analyse_descriptor(&buffer[pos/8],len,todo
						,desctable[d_zeile].sub_table
						,level,channelnr,loop_count);
		    }
		    pos+=len;
		    break;
		case DESC_loop_end:
		    len=desc_length-pos;
		    analyse_descriptor(&buffer[pos/8],len,todo
						,desctable[d_zeile].sub_table
						,level,channelnr,loop_count);
		    pos+=len;
		    break;
		case DESC_loop_end_memcpy:
                    len=desc_length-pos;
                    for(j=0;j<todo->desc_encoding_size;j++) {
                        if(desctable[d_zeile].sub_table==(todo->desc_encoding[j].desc_id)) {
                            // length byte
                            *(todo->desc_encoding[j].save_in
                            +todo->desc_encoding[j].save_in_struct_size
                            *loop_count
                            /sizeof(todo->desc_encoding[j].save_in)
                            +todo->encoding[0].save_in_struct_size
                            *channelnr
                            /sizeof(todo->encoding[0].save_in))=len/8;

                            // data
                            memcpy((char*)(todo->desc_encoding[j].save_in
                            +todo->desc_encoding[j].save_in_struct_size
                            *loop_count
                            /sizeof(todo->desc_encoding[j].save_in)
                            +todo->encoding[0].save_in_struct_size
                            *channelnr
                            /sizeof(todo->encoding[0].save_in))+1
                            ,buffer+(pos/8)
                            ,len/8);
                        }
                    }
                    pos+=len;
                    break;
		case LINKAGE:
		    len=desc_length-pos;
		    if (last!=8) {
			subtable=-8;
		    } else {
			subtable=-9;
		    }
		    analyse_descriptor(&buffer[pos/8],len,todo,subtable,level
							,channelnr,loop_count);
		    pos+=len;
		    break;
		case VS_loop:
		    if(buffer[pos]&0x4) {
			subtable=-1;
			analyse_descriptor(&buffer[pos/8],16,todo,subtable
						,level,channelnr,loop_count);
		    }
		    pos+=16;
		    break;
		default:
		    last=getbits(buffer,pos,desctable[d_zeile].bit_count);
		    if(ALL) {
			tprintf(OUTPUT,*level,"%s:: 0x%x\n"
			    ,desctable[d_zeile].name
			    ,getbits(buffer,pos,desctable[d_zeile].bit_count));
		    } else {
			for(j=0;j<todo->desc_encoding_size;j++) {
			    if(descriptor_tag==(todo->desc_encoding[j].desc_id)) {
				if(d_zeile==(todo->desc_encoding[j].d_zeile)) {
				    *(todo->desc_encoding[j].save_in
				    +todo->desc_encoding[j].save_in_struct_size
				    *loop_count
				    /sizeof(todo->desc_encoding[j].save_in)
				    +todo->encoding[0].save_in_struct_size
				    *channelnr
				    /sizeof(todo->encoding[0].save_in)) 
				    =getbits(buffer,pos,desctable[d_zeile].bit_count);
				    LOOP=1;
				    break;
				}
			    }
			}
		    }
		    pos+=desctable[d_zeile].bit_count;
	    }
	}
	if(LOOP) {
	    loop_count++;
	}
	if(DEBUG) {
	    printf("channelnr: %d\n",channelnr);
	}
    }
    (*level)--;
    if(ALL) {
	tprintf(OUTPUT,*level,"<\\%s=0x%x> <hat %x bytes>\n"
			,descriptorId[tables.stable].name,descriptor_tag,pos/8);
    }
    return pos;
}

int descriptoren(unsigned char *buffer,unsigned int bufferlength,LIST *todo,unsigned int *level,int channelnr,FILE *output) 
{
    unsigned int last_descriptor_tag;
    unsigned int descriptor_tag;
    unsigned int desc_length;
    unsigned int pos;
    int loop_count[0x200];	//descriptor wiederholung pro tabelle

    memset(loop_count,0,sizeof(loop_count));
    LOOP=0;
    OUTPUT=output;
    (*level)++;
    last_descriptor_tag=getbits(&buffer[0/8],0,8);
    for(pos=0;pos<bufferlength;) {
        descriptor_tag=getbits(&buffer[pos/8],0,8);
        desc_length=getbits(&buffer[pos/8],8,8)*8;
        pos+=16;
	LOOP=0;
        pos+=analyse_descriptor(&buffer[pos/8],desc_length,todo,descriptor_tag
						   ,level,channelnr,loop_count[descriptor_tag]);
	if(LOOP==1) {
	    loop_count[descriptor_tag]++;
	}
    }
    (*level)--;
    return pos;
}
