/*
 * Copyright (C) 2002 2003 Sven Schaepe <schaepe@rz.tu-ilmenau.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>

#include "utils.h"
#include "desc.h"
#include "dvbsi.h"

#define DEBUG 0
#define DEBUG2 0
#define INFO 0

FILE *OUTPUT;

typedef struct {
  int table_id;
  const char *name;
  sTABLE *table;
  unsigned int length;
} sTABLE_ID;
		
/* TABLES */

sTABLE defTable[] = {
  { "table_id",                   8, 0 , 0 },
  { "section_syntax_indicator",   1, 0 , 0 },
  { "reserved",                   1, 0 , 0 },
  { "reserved",                   2, 0 , 0 },
  { "section_length",            12, 0 , 0 }
};

sTABLE stdTable[] = {
  { "header",                     0,-11, SUB_TAB },
  { "service_id",                16, 0 , 0 },
  { "reserved",                   2, 0 , 0 },
  { "version_nr",                 5, 0 , 0 },
  { "current_next_indicator",     1, 0 , 0 },
  { "section_nr",                 8, 0 , 0 },
  { "last_section_nr",            8, 0 , 0 }
};

sTABLE pmtTable[] = {
  { "stdTable",                      0,-1 , SUB_TAB },
  { "reserved",                      3, 0 , 0 },
  { "PCR_PID",                      13, 0 , 0 },
  { "reserved",                      4, 0 , 0 },
  { "programm_info_length",         12, 0 , 0 },
  { "DESC",                          0,0x8000, DESC_kl },
  { "loop",                          0,-4 , LOOP_ul },
  { "crc_32",                       32, 0 , 0 }
};

sTABLE pmtlTable[] = {
  { "stream_type",                 8, 0 , 0 },
  { "reserved",                    3, 0 , 0 },
  { "elementary_PID",             13, 0 , 0 },
  { "reserved",                    4, 0 , 0 },
  { "ES_info_length",             12, 0 , 0 },
  { "DESC",                        0,0x8000, DESC_kl }
};

sTABLE nitTable[] = {
  { "stdTable",                      0,-1 , SUB_TAB },
  { "reserved",                      4, 0 , 0 },
  { "network_descriptor_length",    12, 0 , 0 },
  { "DESC",                          0,0x8000, DESC_kl },
  { "reserved",                      4, 0 , 0 },
  { "transport_stream_loop_length", 12, 0 , 0 },
  { "loop",                          0,-5 , LOOP_ul },
  { "crc_32",                       32, 0 , 0 }
};

sTABLE nitlTable[] = {
  { "transport_stream_id",          16, 0 , 0 },
  { "orginal_network_id",           16, 0 , 0 },
  { "reserved_future_use",           4, 0 , 0 },
  { "transport_descriptor_length",  12, 0 , 0 },
  { "DESC",                          0,0x8000, DESC_kl }
};

sTABLE patTable[] = {
  { "stdTable",                   0, -1 , SUB_TAB },
  { "loop",                       0, -3 , LOOP_ul },
  { "crc_32",                     32, 0 , 0 }
};

sTABLE patlTable[] = {
  { "service_id",             	  16, 0 , 0 },
  { "reserved",                    3, 0 , 0 },
  { "network/pmt_pid",            13, 0 , 0 },
};

sTABLE catTable[] = {
  { "header",                     0,-11, SUB_TAB },
  { "reserved",                   18, 0 , 0 },
  { "version_number",              5, 0 , 0 },
  { "current_next_indicator",      1, 0 , 0 },
  { "section_number",              8, 0 , 0 },
  { "last_section_number",         8, 0 , 0 },
  { "DESC1",                       0,0x8000, DESC_ul },
  { "crc_32",                     32, 0 , 0 }
};

sTABLE batTable[] = {
  { "stdTable",                      0,-1 , SUB_TAB },
  { "reserverd_future_use",          4, 0 , 0 },
  { "bouquet_desc_length",          12, 0 , 0 },
  { "DESC",                          0,0x8000, DESC_kl },
  { "reserverd_future_use",          4, 0 , 0 },
  { "transport_stream_loop_length", 12, 0 , 0 },
  { "loop",                          0,-6 , LOOP_ul },
  { "crc_32",                       32, 0 , 0 }
};

sTABLE batlTable[] = {
  { "transport_stream_id",        16, 0 , 0 },
  { "orginal_network_id",         16, 0 , 0 },
  { "reserved_future_use",         4, 0 , 0 },
  { "transport_desc_length",      12, 0 , 0 },
  { "DESC",                        0,0x8000, DESC_kl }
};

sTABLE eitTable[] = {
  { "stdTable",                    0,-1 , SUB_TAB },
  { "transponder_stream_id",      16, 0 , 0 },
  { "orginal_network_id",         16, 0 , 0 },
  { "segment_last_section_nr",     8, 0 , 0 },
  { "last_table_id",               8, 0 , 0 },
  { "loop",                        0,-2 , LOOP_ul },
  { "crc_32",                     32, 0 , 0 }
};

sTABLE eitlTable[] = {
  { "event_id",                   16, 0 , 0 },
  { "start_date",                 16, 0 , 0 },
  { "start_time",                 24, 0 , 0 },
  { "duration",                   24, 0 , 0 },
  { "running_status",              3, 0 , 0 },
  { "free_CA_mode",                1, 0 , 0 },
  { "descriptor_loop_length",     12, 0 , 0 },
  { "DESC",                        0,0x8000, DESC_kl }
};

sTABLE sdtTable[] = {
  { "stdTable",                    0,-1 , SUB_TAB },
  { "orginal_network_id",         16, 0 , 0 },
  { "reserverd_future_use",        8, 0 , 0 },
  { "loop",                        0,-7 , LOOP_ul },
  { "crc_32",                     32, 0 , 0 }
};

sTABLE sdtlTable[] = {
  { "service_id",                 16, 0 , 0 },
  { "reserved_future_use",         6, 0 , 0 },
  { "EIT_schedule_flag",           1, 0 , 0 },
  { "EIT_present_follow_flag",     1, 0 , 0 },
  { "running_status",              3, 0 , 0 },
  { "free_CA_mode",                1, 0 , 0 },
  { "descriptor_loop_length",     12, 0 , 0 },
  { "DESC",                        0,0x8000, DESC_kl }
};

sTABLE rstTable[] = {
  { "header",                     0,-11, SUB_TAB },
  { "loop_1",                      0,-8 , LOOP_kl }
};

sTABLE rstlTable[] = {
  { "transport_stream_id",        16, 0 , 0 },
  { "original_network_id",        16, 0 , 0 },
  { "service_id",                 16, 0 , 0 },
  { "event_id",                   16, 0 , 0 },
  { "reserved_future_use",         5, 0 , 0 },
  { "running_status",              3, 0 , 0 }
};

sTABLE stTable[] = {
//  { "table_id",                   8, 0 , 0 },
//  { "section_syntax_indicator",   1, 0 , 0 },
//  { "reserved",                   1, 0 , 0 },
//  { "reserved",                   2, 0 , 0 },
//  { "section_length",            12, 0 , 0 },
  { "header",                     0,-11, SUB_TAB },
  { "loop_1",                      0,-9 , MEMCPY_kl }
};

sTABLE stlTable[] = {
  { "data_byte",                    8, 0 , 0 }
};

sTABLE totTable[] = {
  { "header",                     0,-11, SUB_TAB },
  { "UTC_date",                   16, 0 , 0 },
  { "UTC_time",                   24, 0 , 0 },
  { "reserved",                    4, 0 , 0 },
  { "descriptor_loop_length",     12, 0 , 0 },
  { "DESC",                        0,0x8000, DESC_kl },
  { "crc_32",                     32, 0 , 0 }
};

sTABLE tdtTable[] = {
  { "header",                     0,-11, SUB_TAB },
  { "UTC_date",                   16, 0 , 0 },
  { "UTC_time",                   24, 0 , 0 },
};

sTABLE ditTable[] = {
  { "header",                     0,-11, SUB_TAB },
  { "transition_flag",             1, 0 , 0 },
  { "reserved_future_use",         7, 0 , 0 }
};

sTABLE ecm_sub7812Table[] = {
  { "seperator",                 16, 0 , 0 },
  { "keynumber",      	          8, 0 , 0 },
  { "unkown",                     8, 0 , 0 },
  { "k1(7-4)",                   32, 0 , 0 },
  { "k1(3-0)",                   32, 0 , 0 },
  { "k2(7-4)",                   32, 0 , 0 },
  { "k2(3-0)",                   32, 0 , 0 },
  { "NANO",                      16, 0 , 0 },
  { "date",                      16, 0 , 0 },
  { "checksum",                   8, 0 , 0 }
};

sTABLE ecmTable[] = {
  { "seperator",                 16, 0 , 0 },
  { "ecm_length",                 8, 0 , 0 },
  { "unkown",                     8, 0 , 0 },
  { "unkown",                     8, 0 , 0 },
  { "unkown",                     8, 0 , 0 },
  { "channel_id",                16, 0 , 0 },
  { "provider 00/10",             8, 0 , 0 },
  { "keynumber",      	          8, 0 , 0 },
  { "keynumber odd/even",         8, 0 , 0 },
  { "length",                     8, 0 , 0 },
  { "ecm_sub",                    0,-12, ECM_SUB_TAB }
};

sTABLE emmTable[] = {
  { "table_id",                   8, 0 , 0 },
  { "sub_table",                  8, 0 , 0 },
  { "length",                     8, 0 , 0 },
  { "address_mode",               8, 0 , 0 },
  { "emm_sub",                    0,-13, EMM_SUB_TAB }
};

sTABLE_ID tableId[] = {
  { 0x8000, "DESC" , 0     , 0                    },  
  { -12   , "ecm_sub7812", ecm_sub7812Table , TABLESIZE(ecm_sub7812Table)  },
  { -11   , "head", defTable , TABLESIZE(defTable)  },
  { -9    , "st_l" , stlTable , TABLESIZE(stlTable)  },
  { -8    , "rst_l", rstlTable, TABLESIZE(rstlTable) },
  { -7    , "sdt_l", sdtlTable, TABLESIZE(sdtlTable) },
  { -6    , "bat_l", batlTable, TABLESIZE(batlTable) },
  { -5    , "nit_l", nitlTable, TABLESIZE(nitlTable) },
  { -4    , "pmt_l", pmtlTable, TABLESIZE(pmtlTable) },
  { -3    , "pat_l", patlTable, TABLESIZE(patlTable) },
  { -2    , "eit_l", eitlTable, TABLESIZE(eitlTable) },
  { -1    , "std"  , stdTable , TABLESIZE(stdTable) },
  { 0x0000, "PAT"  , patTable , TABLESIZE(patTable) },
  { 0x0001, "CAT"  , catTable , TABLESIZE(catTable) },
  { 0x0002, "PMT"  , pmtTable , TABLESIZE(pmtTable) },
  { 0x0040, "NIT_at"  , nitTable , TABLESIZE(nitTable) },
  { 0x0041, "NIT_ot"  , nitTable , TABLESIZE(nitTable) },
  { 0x0042, "SDT_at"  , sdtTable , TABLESIZE(sdtTable) },
  { 0x0046, "SDT_ot"  , sdtTable , TABLESIZE(sdtTable) },
  { 0x004a, "BAT"  , batTable , TABLESIZE(batTable) },
  { 0x004e, "anEIT", eitTable , TABLESIZE(eitTable) },
  { 0x004f, "onEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0050, "asEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0051, "asEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0052, "asEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0053, "asEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0054, "asEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0055, "asEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0056, "asEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0057, "asEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0058, "asEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0059, "asEIT", eitTable , TABLESIZE(eitTable) },
  { 0x005a, "asEIT", eitTable , TABLESIZE(eitTable) },
  { 0x005b, "asEIT", eitTable , TABLESIZE(eitTable) },
  { 0x005c, "asEIT", eitTable , TABLESIZE(eitTable) },
  { 0x005d, "asEIT", eitTable , TABLESIZE(eitTable) },
  { 0x005e, "asEIT", eitTable , TABLESIZE(eitTable) },
  { 0x005f, "asEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0060, "osEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0061, "osEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0062, "osEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0063, "osEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0064, "osEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0065, "osEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0066, "osEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0067, "osEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0068, "osEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0069, "osEIT", eitTable , TABLESIZE(eitTable) },
  { 0x006a, "osEIT", eitTable , TABLESIZE(eitTable) },
  { 0x006b, "osEIT", eitTable , TABLESIZE(eitTable) },
  { 0x006c, "osEIT", eitTable , TABLESIZE(eitTable) },
  { 0x006d, "osEIT", eitTable , TABLESIZE(eitTable) },
  { 0x006e, "osEIT", eitTable , TABLESIZE(eitTable) },
  { 0x006f, "osEIT", eitTable , TABLESIZE(eitTable) },
  { 0x0070, "TDT"  , tdtTable , TABLESIZE(tdtTable) },
  { 0x0071, "RST"  , rstTable , TABLESIZE(rstTable) },
  { 0x0072, "ST"   , stTable  , TABLESIZE(stTable)  },
  { 0x0073, "TOT"  , totTable , TABLESIZE(totTable) },
  { 0x007e, "DIT"  , ditTable , TABLESIZE(ditTable) },
  { 0x0080, "ECM"  , ecmTable , TABLESIZE(ecmTable) },
  { 0x0081, "ECM"  , ecmTable , TABLESIZE(ecmTable) }
  //{ 0x0082, "EMM"  , emmTable , TABLESIZE(emmTable) }
};

unsigned int gettable(int table_id) 
{
    unsigned int zeile;

    for (zeile=0;zeile<TABLESIZE(tableId);zeile++) {
        if (tableId[zeile].table_id==table_id) {
            return zeile;
        }
    }
    return (0xffff);
}

int analyse_table(unsigned char *buffer,unsigned int sectionlength
		    ,int table_id,LIST *todo
		    ,unsigned int *level,int *channel)
{
    sTABLE *table;
    unsigned int len=0;
    unsigned int pos;
    unsigned int length;
    unsigned int zeile,t_zeile;
    int /*channel=0,*/t;
    char channelcount=0;

    if ((zeile=(gettable(table_id)))==0xffff) {    /* unknown table */
        print_hex_dump(buffer,sectionlength/8,level,OUTPUT);
	//sectionlength=0;
        return sectionlength;
    }

    if(DEBUG) {
	printf("table: %x %s\n",tableId[zeile].table_id,tableId[zeile].name);
    }

    table=tableId[zeile].table;

    for(pos=0;pos<sectionlength;) {  
	if(ALL) {
	    tprintf(OUTPUT,*level,"<%s>\n",tableId[zeile].name);
	}
	(*level)++;
	for(t_zeile=0;t_zeile<(tableId[gettable(table_id)].length);t_zeile++) {
	    switch (table[t_zeile].encoding) {
	      case SUB_TAB:
		  (*level)--;
		  {
		  int tmp=0;
		  length=analyse_table(&buffer[pos/8],8,table[t_zeile].sub_table
							   ,todo,level,&tmp);
		  }
		  pos+=length;
		  (*level)++;
		  break;
	      case MEMCPY_kl: 
		   len*=8;
		   if(channelcount) {
		       channelcount=0;
		       (*channel)++;
		   }
		   for(t=0;t<todo->encoding_size;t++) {
			if(table_id==(todo->encoding[t].table_id)) {
			    if(t_zeile==(todo->encoding[t].t_zeile)) { 
				 memcpy(&buffer[pos/8],todo->encoding[t].save_in
					 +todo->encoding[t].save_in_struct_size
					 *(*channel)
					 /sizeof(todo->encoding[t].save_in),len);
				 channelcount=1;
			    }
			}
		   }
		   pos+=len;
		   break;
	      case LOOP_kl: 
		  len*=8;
		  if(channelcount) {
		      channelcount=0;
		      (*channel)++;
		  }
		  length=analyse_table(&buffer[pos/8],len,table[t_zeile].sub_table
							   ,todo,level,channel);
		  if(length!=len) {
		      fprintf(stderr,"oops at LOOP_kl: length=%d != len=%d !!\n",
								    length,len);
		  }
		  pos+=length;
		  break;
	      case LOOP_ul:
		  len=sectionlength-pos-32; //32bit CRC
		  if(channelcount) {
		      channelcount=0;
		      (*channel)++;
		  }
		  length=analyse_table(&buffer[pos/8],len,table[t_zeile].sub_table
							   ,todo,level,channel);
		  if(length!=len) {
		      fprintf(stderr,"oops at LOOP_ul: length=%d != len=%d !!\n",
								    length,len);
		  }
		  pos+=length;
		  break;
	      case DESC_ul: 
		  len=sectionlength-pos-32; //32bit CRC
		  length=descriptoren(&buffer[pos/8],len,todo,level,*channel
								    ,OUTPUT);
		  if(length!=len) {
		      fprintf(stderr,"oops at DESC_ul: length=%d != len=%d !!\n",
								    length,len);
		  }
		  pos+=length;
		  break;
	      case DESC_kl: 
		  len*=8;
		  length=descriptoren(&buffer[pos/8],len,todo,level,*channel
								    ,OUTPUT);
		  if(length!=len) {
		      fprintf(stderr,"oops at LOOP_kl: length=%d != len=%d !!\n",
								    length,len);
		  }
		  pos+=length;
		  break;
	      default:
		  len=getbits(buffer,pos,table[t_zeile].bit_count);
		  if (ALL) {  
		      tprintf(OUTPUT,*level,"%s: 0x%x %d\n",table[t_zeile].name,len,len);
		  } else {
		      if(todo->filter!=NULL && !todo->filter->ok) {
			  if(strcmp(table[t_zeile].name,"service_id")==0) {
			      if(getbits(buffer,pos,table[t_zeile].bit_count)
			      				==todo->filter->value) {
				  todo->filter->ok=1;
			      } 
			  }
		      } 
		      if(todo->filter==NULL || todo->filter->ok) {
			  for (t=0;t<todo->encoding_size;t++) {
			      if (table_id==(todo->encoding[t].table_id)) {
				  if (t_zeile==(todo->encoding[t].t_zeile)) { 
				      *(todo->encoding[t].save_in
				      +todo->encoding[t].save_in_struct_size
				      *(*channel)
				      /sizeof(todo->encoding[t].save_in))
				      =getbits(buffer,pos,table[t_zeile].bit_count);
				      channelcount=1;
				  }
			      }
			  }	
		      }
		  }
		  pos+=table[t_zeile].bit_count;
	      } 
	  }  
	  if(ALL) {
	      tprintf(OUTPUT,*level,"<\\%s>\n",tableId[zeile].name);
	  } else {
	      if(channelcount) {
		  channelcount=0;
		  (*channel)++;
	      }
	  }
	  (*level)--;
    }
    if(todo->channels<=(*channel)) {
	todo->channels=*channel;
    }
    if(DEBUG) {
	printf("channels: %d\n",*channel);
    }
    return pos;
}

unsigned char analyse_table_head(unsigned char *buffer) 
{
  unsigned char last_section_number=0;

    if (getbits(buffer,8,1)==1) 
      last_section_number=getbits(buffer,56,8); 
    return last_section_number;
}

unsigned char get_si_stream(unsigned char *buffer,unsigned int buffer_size
					    ,LIST *todo,char all,FILE *output)
{
    int i=0;
    int table_id;
    int channel=0;
    unsigned int level=0;
    unsigned char loop=0;

    ALL=all;
    OUTPUT=output;

    table_id=getbits(buffer,0,8);

    /* setze filter check auf 0 */
    if(todo->filter!=NULL) {
	todo->filter->ok=0;
    }

    analyse_table(&buffer[0],buffer_size*8,table_id,todo,&level,&channel);

    if(ALL) {
	printf("%x packete und last_section_num = %x\n",i,loop);
    }

    return loop;
}
